from wfs20.error import WFSError

import warnings
import importlib.util

try:
	from osgeo import ogr, osr
	_FieldTypes = {
	int:ogr.OFTInteger64,
	float:ogr.OFTReal,
	str:ogr.OFTString
	}
except ModuleNotFoundError:
	warnings.warn("osgeo package not installed. Writing to shapefile is not available.",ImportWarning)

def _WriteShapeFile(reader,out):
	"""
	"""
	if not importlib.util.find_spec("osgeo"):
		raise ModuleNotFoundError("Cannot execute function as osgeo is not installed.")
	Driver = ogr.GetDriverByName('ESRI Shapefile')

	srs = osr.SpatialReference()
	srs.ImportFromEPSG(28992)

	dst = Driver.CreateDataSource(f"{out}\\{reader.Keyword}.shp")
	Layer = dst.CreateLayer(reader.Keyword,srs)

	# Create Fields of Layer
	FieldTypes = reader.LayerMeta.FieldTypes
	LinkTable = reader.LayerMeta.LinkTable

	for header,t in FieldTypes.items():
		field = ogr.FieldDefn(
			LinkTable[header],
			_FieldTypes[t]
			)
		if t == str:
			field.SetWidth(100)
		Layer.CreateField(field)

	# Create and add Features to Layer
	for f in reader.Features:
		Feature = ogr.Feature(Layer.GetLayerDefn())
		Geometry = ogr.CreateGeometryFromGML(f.Geometry.decode())
		for header,v in f.Fields.items():
			Feature.SetField(
				LinkTable[header],
				v
				)
		Feature.SetGeometry(Geometry)
		Layer.CreateFeature(Feature)

	# Garbage collection by force
	FieldTypes = None
	LinkTable = None
	reader = None
	Feature = None
	Layer = None
	dst = None


from wfs20.request import parse_qsl, GetResponse 
from wfs20.util import _BuildResonseMeta

def _ServiceReader(url,timeout):
	"""
	Service Reader
	"""
	r = GetResponse(url,timeout=timeout)
	return r

class DataReader:
	"""
	Response reader of a geospatial data request

	Parameters
	----------
	url: str
		request url for geospatial data
	keyword: str
		Designation of the requested layer
	method: str
		Request method, either 'GET' or 'POST'
	data: str
		Params in xml format

	Returns
	-------
	Reader object
	"""
	def __init__(
		self,
		url,
		keyword,
		method="GET",
		data=None
		):

		# General stuff
		self.URL = url
		self.Keyword = keyword
		self.RequestMethod = method
		self.RequestData = data

		# substance
		_BuildResonseMeta(self, GetResponse(self.URL, timeout=30, method=method, data=data), self.Keyword)

	def __repr__(self):
		return super().__repr__()

	def __iadd__(self,other):
		if isinstance(self, other.__class__):
			self.Features += other.Features
			self.LayerMeta |= other.LayerMeta
			return self
		else:
			raise TypeError(f"unsupported operand type(s) for +=: '{self.__class__}' and '{other.__class__}'")


def execute_query(conn, query):
    cur = conn.cursor()
    try:
        cur.execute(query)
        conn.commit()
    except Exception as e:
        print(f"The error '{e}' occurred")
    cur.close()

def execute_read_query(conn, query):
    cur = conn.cursor()
    r = None
    try:
        cur.execute(query)
        r = cur.fetchall()
        return r
    except Exception as e:
        print(f"The error '{e}' occurred")
    cur.close()


from wfs20 import __path__
from wfs20.io.util import execute_query

import os
import sys
import sqlite3
from osgeo import osr

# 0,5,6 are hard to define.
_OrienTable = {
	0:"xy",
	1:"yx",
	2:"yx",
	3:"xy",
	4:"xy",
	5:"xy",
	6:"xy"	
}

def _CreateAxisOrderDBASE():
	"""
	"""
	# Some locations
	pyloc = os.path.dirname(sys.executable)
	# database connections
	proj = sqlite3.connect(f"{pyloc}\\Lib\\site-packages\\osgeo\\data\\proj\\proj.db")
	conn = sqlite3.connect(f"{__path__[0]}\\data\\axisorder.db")
	# set cursor
	proj_cur = proj.cursor()
	create_table = """\
CREATE TABLE IF NOT EXISTS axisorder (
'auth' TEXT NOT NULL,
'code' INTEGER PRIMARY KEY AUTOINCREMENT,
'order' TEXT NOT NULL,
'reference' TEXT
);
"""
	execute_query(conn, create_table)
	for crs_type in ["projected_crs","geodetic_crs","vertical_crs","compound_crs"]:
		for code in tuple(proj_cur.execute(f"SELECT code FROM {crs_type} WHERE auth_name = 'EPSG';")):
			srs = osr.SpatialReference()
			srs.ImportFromEPSG(code[0])
			url = f"http://epsg.io/{code[0]}"
			order = _OrienTable[srs.GetAxisOrientation(None,0)]
			print(srs.GetAxisOrientation(None,0))
			add_to_table = f"""\
INSERT INTO
  	axisorder ('auth','code','order','reference')
VALUES
  	("EPSG",{code[0]},'{order}','{url}')
"""
			execute_query(conn, add_to_table)
			print(f"Succesfully added EPSG:{code[0]}")
			# clean up the srs
			srs = None
	# close all connections and cursors
	conn.close()
	proj_cur.close()
	proj.close()

if __name__ == "__main__":
	_CreateAxisOrderDBASE()


from wfs20.crs import CRS
from wfs20.error import WFSInternalError
from wfs20.io import _WriteShapeFile, _WriteToGML
from wfs20.reader import _ServiceReader, DataReader
from wfs20.request import _ServiceURL, CreateGetRequest
from wfs20.util import _BuildServiceMeta

import sys

class WebFeatureService:
	"""
	WebFeatureService

	Parameters
	----------

	url: str
		Service url
	version: str
		WebFeatureService version
	
	Returns
	-------
	WFS reader object: wfs20.WebFeatureService
		Object containing the metadata of the service
	"""
	def __init__(self,url,version="2.0.0"):
		self.url = url
		self.version = version
		self.ServiceURL = _ServiceURL(self.url,version)

		# Substance
		_BuildServiceMeta(self, _ServiceReader(self.ServiceURL,timeout=30))

		# Declarations
		self.DataReader = None

	def __repr__(self):
		return f"<wfs20.WebFeatureService object ({self.url})>"

	def RequestData(
		self,
		featuretype,
		bbox,
		epsg
		):
		"""
		Request spatial data from the WebFeatureService

		Parameters
		----------
		featuretype: str
			Layer to be requested, mostly in the format of 'xxx:xxx'
		bbox: list or tuple
			Bounding box wherein the spatial data lies that is requested,
			e.g. (x1,y1,x2,y2)
		epsg: int
			The projection code of the requested data and the bounding box 
			according to EPSG, e.g. 4326 (WGS84)

		Returns
		-------
		Reader object: wfs20.reader.DataReader
			Contains the requested data
		"""
		if featuretype not in self.FeatureTypes:
			raise WFSInternalError(
				"Request Error", 
				f"<{featuretype}> not in list of available featuretypes (see <class>.FeatureTypes)"
				)
		crs = CRS.from_epsg(epsg)
		if not crs in self.FeatureTypeMeta[featuretype].CRS:
			raise WFSInternalError(
				"Request Error", 
				f"<{epsg}> not in list of available projections (see <class>.FeatureTypeMeta[<id>].CRS)"
				)
		url = CreateGetRequest(
				self.url,
				self.version,
				featuretype,
				bbox,
				crs
				)
		keyword = self.FeatureTypeMeta[featuretype].Title
		self.DataReader = DataReader(url,keyword)
		return self.DataReader

	def ToFile(
		self,
		out,
		format="shp"
		):
		"""
		Write geospatial data held in reader to file

		Parameters
		----------
		out: str
			path of where the file should be written to
		format: str
			format of the file, following options are available:
				- gml
				- shp
		"""
		if self.DataReader == None or not self.DataReader.Features:
			raise WFSInternalError("Writing to file","No features collected from WebFeatureService")
		if format == "shp":
			_WriteShapeFile(self.DataReader,out)
		elif format == "gml":
			_WriteToGML(self.DataReader,out)
		else:
			raise ValueError(f"Incorrect format => {format}")


from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
	name="wfs20",
	version="0.2.1",
	description="Small library to request geospatial data (WFS)",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/B-Dalmijn/WFS20",
	author="Brendan Wybo Dalmijn",
	author_email="brencodeert@outlook.com",
	license="BSD",
	classifiers=[
	# How mature is this project? Common values are
	#   3 - Alpha
	#   4 - Beta
	#   5 - Production/Stable
	'Development Status :: 4 - Beta',

	# Indicate who your project is intended for
	'Intended Audience :: End Users/Desktop',
	'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',

    # Topic of the package
    'Topic :: Scientific/Engineering :: GIS',

	# Pick your license as you wish (should match "license" above)
	'License :: OSI Approved :: BSD License',

	# Language
	'Natural Language :: English',

	# OS
	'Operating System :: OS Independent',

	# Specify the Python versions you support here. In particular, ensure
	# that you indicate whether you support Python 2, Python 3 or both.
	'Programming Language :: Python :: 3',
	'Programming Language :: Python :: 3.6',
	'Programming Language :: Python :: 3.7',
	'Programming Language :: Python :: 3.8',
	'Programming Language :: Python :: 3.9',
	],
	keywords="wfs WebFeatureService shapefile gml request",
	project_urls={
	'Source': 'https://github.com/B-Dalmijn/WFS20/',
	'Tracker': 'https://github.com/B-Dalmijn/WFS20/issues',
	},
	package_dir={"":"src"},
	packages=find_packages(where="src"),
	install_requires=[
	"requests",
	"lxml"
	],
	python_requires='>=3.6',
	package_data={
	"wfs20": ["data\\axisorder.db"],
	}
	)